@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2021 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2022 the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -71,7 +71,6 @@
 #endif
 
 #include <stdarg.h>
-#include <stdbool.h>
 #include <signal.h>
 #include <time.h>
 #include <errno.h>
@@ -156,22 +155,13 @@ typedef int off_t;
 #define setlocale(locale, val)	/* nothing */
 #endif /* HAVE_SETLOCALE */
 
-#if HAVE_MEMCPY_ULONG
-extern char *memcpy_ulong(char *dest, const char *src, unsigned long l);
-#define memcpy memcpy_ulong
-#endif
-#if HAVE_MEMSET_ULONG
-extern void *memset_ulong(void *dest, int val, unsigned long l);
-#define memset memset_ulong
-#endif
-
 #ifdef HAVE_FWRITE_UNLOCKED
 #define fwrite	fwrite_unlocked
 #endif /* HAVE_FWRITE_UNLOCKED */
 
-#if defined(__DJGPP__) || defined(__EMX__) || defined(__MINGW32__)
+#if defined(__MINGW32__)
 #include "nonposix.h"
-#endif /* defined(__DJGPP__) || defined(__EMX__) || defined(__MINGW32__) */
+#endif /* defined(__MINGW32__) */
 
 /* use this as lintwarn("...")
    this is a hack but it gives us the right semantics */
@@ -213,6 +203,8 @@ typedef struct Regexp {
 
 #include "gawkapi.h"
 
+#include "floatmagic.h"
+
 /* Stuff for losing systems. */
 #if !defined(HAVE_STRTOD)
 extern double gawk_strtod();
@@ -258,6 +250,7 @@ typedef enum nodevals {
 	Node_var,		/* scalar variable, lnode is value */
 	Node_var_array,		/* array is ptr to elements, table_size num of eles */
 	Node_var_new,		/* newly created variable, may become an array */
+	Node_elem_new,		/* newly created array element, may become a subarray */
 	Node_param_list,	/* lnode is a variable, rnode is more list */
 	Node_func,		/* lnode is param. list, rnode is body */
 	Node_ext_func,		/* extension function, code_ptr is builtin code */
@@ -436,6 +429,10 @@ typedef struct exp_node {
 		 * is a hint to indicate that an integer array optimization may be
 		 * used when this value appears as a subscript.
 		 *
+		 * The BOOL flag indicates that this number should be converted to True
+		 * or False by extensions that interchange data with other languages,
+		 * via JSON, XML or some other serialization mechanism.
+		 *
 		 * We hope that the rest of the flags are self-explanatory. :-)
 		 */
 		MALLOC	= 0x0001,       /* stptr can be free'd, i.e. not a field node pointing into a shared buffer */
@@ -445,24 +442,25 @@ typedef struct exp_node {
 		NUMBER	= 0x0010,       /* assigned as number */
 		USER_INPUT = 0x0020,    /* user input: if NUMERIC then
 					 * a NUMBER */
-		INTLSTR	= 0x0040,       /* use localized version */
-		NUMINT	= 0x0080,       /* numeric value is an integer */
-		INTIND	= 0x0100,	/* integral value is array index;
+		BOOLVAL = 0x0040,	/* this is a boolean value */
+		INTLSTR	= 0x0080,       /* use localized version */
+		NUMINT	= 0x0100,       /* numeric value is an integer */
+		INTIND	= 0x0200,	/* integral value is array index;
 					 * lazy conversion to string.
 					 */
-		WSTRCUR	= 0x0200,       /* wide str value is current */
-		MPFN	= 0x0400,       /* arbitrary-precision floating-point number */
-		MPZN	= 0x0800,       /* arbitrary-precision integer */
-		NO_EXT_SET = 0x1000,    /* extension cannot set a value for this variable */
-		NULL_FIELD = 0x2000,    /* this is the null field */
+		WSTRCUR	= 0x0400,	/* wide str value is current */
+		MPFN	= 0x0800,	/* arbitrary-precision floating-point number */
+		MPZN	= 0x01000,	/* arbitrary-precision integer */
+		NO_EXT_SET = 0x02000,	/* extension cannot set a value for this variable */
+		NULL_FIELD = 0x04000,	/* this is the null field */
 
 	/* type = Node_var_array */
-		ARRAYMAXED	= 0x4000,       /* array is at max size */
-		HALFHAT		= 0x8000,       /* half-capacity Hashed Array Tree;
+		ARRAYMAXED	= 0x08000,	/* array is at max size */
+		HALFHAT		= 0x010000,	/* half-capacity Hashed Array Tree;
 						 * See cint_array.c */
-		XARRAY		= 0x10000,
-		NUMCONSTSTR	= 0x20000,	/* have string value for numeric constant */
-		REGEX           = 0x40000,	/* this is a typed regex */
+		XARRAY		= 0x020000,
+		NUMCONSTSTR	= 0x040000,	/* have string value for numeric constant */
+		REGEX           = 0x080000,	/* this is a typed regex */
 	} flags;
 	long valref;
 } NODE;
@@ -1192,6 +1190,7 @@ extern enum do_flag_values {
 extern bool do_optimize;
 extern int use_lc_numeric;
 extern int exit_val;
+extern bool using_persistent_malloc;
 
 #ifdef NO_LINT
 #define do_lint 0
@@ -1376,8 +1375,8 @@ extern void r_freeblock(void *, int id);
 #define		ALREADY_MALLOCED	2
 #define		ELIDE_BACK_NL		4
 
-#define	cant_happen()	r_fatal("internal error line %d, file: %s", \
-				__LINE__, __FILE__)
+#define	cant_happen(format, ...)	r_fatal("internal error: file %s, line %d: " format, \
+				__FILE__, __LINE__, __VA_ARGS__)
 
 #define	emalloc(var,ty,x,str)	(void) (var = (ty) emalloc_real((size_t)(x), str, #var, __FILE__, __LINE__))
 #define	ezalloc(var,ty,x,str)	(void) (var = (ty) ezalloc_real((size_t)(x), str, #var, __FILE__, __LINE__))
@@ -1439,6 +1438,7 @@ extern NODE *do_asorti(int nargs);
 extern unsigned long (*hash)(const char *s, size_t len, unsigned long hsize, size_t *code);
 extern void init_env_array(NODE *env_node);
 extern void init_argv_array(NODE *argv_node, NODE *shadow_node);
+extern NODE *new_array_element(void);
 /* awkgram.c */
 extern NODE *variable(int location, char *name, NODETYPE type);
 extern int parse_program(INSTRUCTION **pcode, bool from_eval);
@@ -1464,6 +1464,7 @@ extern bool is_identchar(int c);
 extern NODE *make_regnode(NODETYPE type, NODE *exp);
 extern bool validate_qualified_name(char *token);
 /* builtin.c */
+extern void efflush(FILE *fp, const char *from, struct redirect *rp);
 extern double double_to_int(double d);
 extern NODE *do_exp(int nargs);
 extern NODE *do_fflush(int nargs);
@@ -1513,6 +1514,9 @@ extern int strncasecmpmbs(const unsigned char *,
 			  const unsigned char *, size_t);
 extern int sanitize_exit_status(int status);
 extern void check_symtab_functab(NODE *dest, const char *fname, const char *msg);
+extern NODE *do_mkbool(int nargs);
+extern void check_exact_args(int nargs, const char *fname, int count);
+extern void check_args_min_max(int nargs, const char *fname, int min, int max);
 /* debug.c */
 extern void init_debug(void);
 extern int debug_prog(INSTRUCTION *pc);
@@ -1549,6 +1553,7 @@ extern STACK_ITEM *grow_stack(void);
 extern void dump_fcall_stack(FILE *fp);
 extern int register_exec_hook(Func_pre_exec preh, Func_post_exec posth);
 extern NODE **r_get_field(NODE *n, Func_ptr *assign, bool reference);
+extern NODE *elem_new_to_scalar(NODE *n);
 /* ext.c */
 extern NODE *do_ext(int nargs);
 void load_ext(const char *lib_name);	/* temporary */
@@ -1586,6 +1591,15 @@ typedef enum {
 extern field_sep_type current_field_sep(void);
 extern const char *current_field_sep_str(void);
 
+typedef enum {
+	SCALAR_EQ,
+	SCALAR_NEQ,
+	SCALAR_LT,
+	SCALAR_LE,
+	SCALAR_GT,
+	SCALAR_GE,
+} scalar_cmp_t;
+
 /* gawkapi.c: */
 extern gawk_api_t api_impl;
 extern void init_ext_api(void);
@@ -1606,6 +1620,7 @@ extern int os_isreadable(const awk_input_buf_t *iobuf, bool *isdir);
 extern int os_is_setuid(void);
 extern int os_setbinmode(int fd, int mode);
 extern void os_restore_mode(int fd);
+extern void os_maybe_set_errno(void);
 extern size_t optimal_bufsize(int fd, struct stat *sbuf);
 extern int ispath(const char *file);
 extern int isdirpunct(int c);
@@ -1642,6 +1657,7 @@ extern bool is_non_fatal_redirect(const char *str, size_t len);
 extern void ignore_sigpipe(void);
 extern void set_sigpipe_to_default(void);
 extern bool non_fatal_flush_std_file(FILE *fp);
+extern size_t gawk_fwrite(const void *buf, size_t size, size_t count, FILE *fp, void *opaque);
 
 /* main.c */
 extern int arg_assign(char *arg, bool initing);
@@ -1685,6 +1701,9 @@ extern NODE *mpg_node(unsigned int);
 extern const char *mpg_fmt(const char *, ...);
 extern int mpg_strtoui(mpz_ptr, char *, size_t, char **, int);
 extern void mpg_zero(NODE *n);
+extern void *mpfr_mem_alloc(size_t alloc_size);
+extern void *mpfr_mem_realloc(void *ptr, size_t old_size, size_t new_size);
+extern void mpfr_mem_free(void *ptr, size_t size);
 #endif
 /* msg.c */
 extern void gawk_exit(int status);
@@ -1715,6 +1734,7 @@ extern NODE *r_force_number(NODE *n);
 extern NODE *r_format_val(const char *format, int index, NODE *s);
 extern NODE *r_dupnode(NODE *n);
 extern NODE *make_str_node(const char *s, size_t len, int flags);
+extern NODE *make_bool_node(bool value);
 extern NODE *make_typed_regex(const char *re, size_t len);
 extern void *more_blocks(int id);
 extern int parse_escape(const char **string_ptr);
@@ -1828,7 +1848,8 @@ POP_ARRAY(bool check_for_untyped)
 	NODE *t = POP();
 	static bool warned = false;
 
-	if (do_lint && ! warned && check_for_untyped && t->type == Node_var_new) {
+	if (do_lint && ! warned && check_for_untyped
+	    && (t->type == Node_var_new || t->type == Node_elem_new)) {
 		warned = true;
 		lintwarn(_("behavior of `for' loop on untyped variable is not defined by POSIX"));
 	}
@@ -1855,6 +1876,8 @@ POP_SCALAR()
 
 	if (t->type == Node_var_array)
 		fatal(_("attempt to use array `%s' in a scalar context"), array_vname(t));
+	else if (t->type == Node_elem_new)
+		t = elem_new_to_scalar(t);
 
 	return t;
 }
@@ -1868,6 +1891,10 @@ TOP_SCALAR()
 
 	if (t->type == Node_var_array)
 		fatal(_("attempt to use array `%s' in a scalar context"), array_vname(t));
+	else if (t->type == Node_elem_new) {
+		t = elem_new_to_scalar(t);	// fix it up
+		REPLACE(t);			// put it back on the stack
+	}
 
 	return t;
 }
@@ -1921,6 +1948,13 @@ dupnode(NODE *n)
 static inline NODE *
 force_string_fmt(NODE *s, const char *fmtstr, int fmtidx)
 {
+	if (s->type == Node_elem_new) {
+		s->type = Node_val;
+		s->flags &= ~NUMBER;
+
+		return s;
+	}
+
 	if ((s->flags & STRCUR) != 0
 		&& (s->stfmt == STFMT_UNUSED || (s->stfmt == fmtidx
 #ifdef HAVE_MPFR
