@@ -3,7 +3,8 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2017 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2022,
+ * the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -32,16 +33,25 @@
  * valref 1, that effectively means that this is an assignment like "$n = $n",
  * so a no-op, other than triggering $0 reconstitution.
  */
-#define UNFIELD(l, r) \
-{ \
-	/* if was a field, turn it into a var */ \
-	if ((r->flags & MALLOC) != 0 || r->valref == 1) { \
-		l = r; \
-	} else { \
-		l = dupnode(r); \
-		DEREF(r); \
-	} \
+
+// not a macro so we can step into it with a debugger
+#ifndef UNFIELD_DEFINED
+#define UNFIELD_DEFINED 1
+static inline void
+unfield(NODE **l, NODE **r)
+{
+	/* if was a field, turn it into a var */
+	if (((*r)->flags & MALLOC) != 0 || (*r)->valref == 1) {
+		(*l) = (*r);
+	} else {
+		(*l) = dupnode(*r);
+		DEREF(*r);
+	}
 }
+
+#define UNFIELD(l, r)	unfield(& (l), & (r))
+#endif
+
 int
 r_interpret(INSTRUCTION *code)
 {
@@ -57,6 +67,7 @@ r_interpret(INSTRUCTION *code)
 	Regexp *rp;
 	NODE *set_array = NULL;	/* array with a post-assignment routine */
 	NODE *set_idx = NULL;	/* the index of the array element */
+	bool in_indirect_call = false;
 
 
 /* array subscript */
@@ -90,9 +101,21 @@ top:
 		}
 #endif
 
-		switch ((op = pc->opcode)) {
+		op = pc->opcode;
+		if (do_itrace) {
+			fprintf(stderr, "+ %s\n", opcode2str(op));
+			fflush(stderr);
+		}
+
+		switch (op) {
 		case Op_rule:
-			currule = pc->in_rule;   /* for sole use in Op_K_next, Op_K_nextfile, Op_K_getline */
+			currule = pc->in_rule;   /* for use in Op_K_next, Op_K_nextfile, Op_K_getline */
+			// 8/2020: See node BEGINFILE/ENDFILE in the manual.  We clear the record
+			// since conceptually we are before reading a new record from the
+			// upcoming file but haven't read it yet.
+			if (currule == BEGINFILE)
+				set_record("", 0, NULL);
+
 			/* fall through */
 		case Op_func:
 			source = pc->source_file;
@@ -101,6 +124,7 @@ top:
 		case Op_atexit:
 		{
 			bool stdio_problem = false;
+			bool got_EPIPE = false;
 
 			/* avoid false source indications */
 			source = NULL;
@@ -116,7 +140,7 @@ top:
 			 * and pipes, in that it doesn't affect their exit status.
 			 * So we no longer do either.
 			 */
-			(void) close_io(& stdio_problem);
+			(void) close_io(& stdio_problem, & got_EPIPE);
 			/*
 			 * However, we do want to exit non-zero if there was a problem
 			 * with stdout/stderr, so we reinstate a slightly different
@@ -126,6 +150,9 @@ top:
 				exit_val = 1;
 
 			close_extensions();
+
+			if (got_EPIPE)
+				die_via_sigpipe();
 		}
 			break;
 
@@ -142,7 +169,10 @@ top:
 				orig = m->stptr;
 				trans = dgettext(TEXTDOMAIN, orig);
 				m->stptr[m->stlen] = save;
-				m = make_string(trans, strlen(trans));
+				if (trans != orig)	// got a translation
+					m = make_string(trans, strlen(trans));
+				else
+					UPREF(m);
 			} else
 				UPREF(m);
 			PUSH(m);
@@ -182,18 +212,42 @@ top:
 
 			case Node_var_new:
 uninitialized_scalar:
-				if (op != Op_push_arg_untyped) {
-					/* convert untyped to scalar */
-					m->type = Node_var;
-					m->var_value = dupnode(Nnull_string);
-				}
 				if (do_lint)
 					lintwarn(isparam ?
 						_("reference to uninitialized argument `%s'") :
 						_("reference to uninitialized variable `%s'"),
 								save_symbol->vname);
-				if (op != Op_push_arg_untyped)
+
+				if (op != Op_push_arg_untyped) {
+					// convert very original untyped to scalar
+					m->type = Node_var;
+					m->var_value = dupnode(Nnull_string);
+
+					// set up local param by value
 					m = dupnode(Nnull_string);
+				}
+
+				UPREF(m);
+				PUSH(m);
+				break;
+
+			case Node_elem_new:
+				if (do_lint)
+					lintwarn(isparam ?
+						_("reference to uninitialized argument `%s'") :
+						_("reference to uninitialized variable `%s'"),
+								save_symbol->vname);
+
+				if (op != Op_push_arg_untyped) {
+					// convert very original untyped to scalar
+					m->type = Node_var;
+					m->var_value = dupnode(Nnull_string);
+
+					// set up local param by value
+					DEREF(m);
+					m = dupnode(Nnull_string);
+				}
+
 				PUSH(m);
 				break;
 
@@ -206,7 +260,7 @@ uninitialized_scalar:
 				break;
 
 			default:
-				cant_happen();
+				cant_happen("unexpected parameter type %s", nodetype2str(m->type));
 			}
 		}
 			break;
@@ -234,21 +288,32 @@ uninitialized_scalar:
 
 		case Op_subscript:
 			t2 = mk_sub(pc->sub_count);
-			t1 = POP_ARRAY();
+			t1 = POP_ARRAY(false);
 
-			if (do_lint && in_array(t1, t2) == NULL) {
+			if (in_array(t1, t2) == NULL) {
 				t2 = force_string(t2);
-				lintwarn(_("reference to uninitialized element `%s[\"%.*s\"]'"),
-					array_vname(t1), (int) t2->stlen, t2->stptr);
-				if (t2->stlen == 0)
-					lintwarn(_("subscript of array `%s' is null string"), array_vname(t1));
+
+				if (t1 == func_table) {
+					fatal(_("reference to uninitialized element `%s[\"%.*s\"] is not allowed'"),
+						"FUNCTAB", (int) t2->stlen, t2->stptr);
+				} else if (t1 == symbol_table) {
+					fatal(_("reference to uninitialized element `%s[\"%.*s\"] is not allowed'"),
+						"SYMTAB", (int) t2->stlen, t2->stptr);
+				} else if (do_lint) {
+					lintwarn(_("reference to uninitialized element `%s[\"%.*s\"]'"),
+						array_vname(t1), (int) t2->stlen, t2->stptr);
+					if (t2->stlen == 0)
+						lintwarn(_("subscript of array `%s' is null string"), array_vname(t1));
+				}
 			}
 
+			// continue the regular processing
+
 			/* for FUNCTAB, get the name as the element value */
 			if (t1 == func_table) {
 				static bool warned = false;
 
-				if (do_lint && ! warned) {
+				if (do_lint_extensions && ! warned) {
 					warned = true;
 					lintwarn(_("FUNCTAB is a gawk extension"));
 				}
@@ -266,48 +331,70 @@ uninitialized_scalar:
 			if (t1 == symbol_table) {
 				static bool warned = false;
 
-				if (do_lint && ! warned) {
+				if (do_lint_extensions && ! warned) {
 					warned = true;
 					lintwarn(_("SYMTAB is a gawk extension"));
 				}
 				if (r->type == Node_var)
 					r = r->var_value;
+				else if (r->type == Node_var_new) {
+					// variable may exist but have never been set.
+					r->var_value = dupnode(Nnull_string);
+					r = r->var_value;
+				}
 			}
 
-			if (r->type == Node_val)
+			if (r->type == Node_val || r->type == Node_elem_new)
 				UPREF(r);
 			PUSH(r);
 			break;
 
 		case Op_sub_array:
 			t2 = mk_sub(pc->sub_count);
-			t1 = POP_ARRAY();
+			t1 = POP_ARRAY(false);
 			r = in_array(t1, t2);
+
+			if (r == NULL) {
+				t2 = force_string(t2);
+
+				if (t1 == func_table) {
+					fatal(_("reference to uninitialized element `%s[\"%.*s\"] is not allowed'"),
+						"FUNCTAB", (int) t2->stlen, t2->stptr);
+				} else if (t1 == symbol_table) {
+					fatal(_("reference to uninitialized element `%s[\"%.*s\"] is not allowed'"),
+						"SYMTAB", (int) t2->stlen, t2->stptr);
+				} else if (do_lint) {
+					lintwarn(_("reference to uninitialized element `%s[\"%.*s\"]'"),
+						array_vname(t1), (int) t2->stlen, t2->stptr);
+					if (t2->stlen == 0)
+						lintwarn(_("subscript of array `%s' is null string"), array_vname(t1));
+				}
+			}
+
 			if (r == NULL) {
 				r = make_array();
 				r->parent_array = t1;
-				lhs = assoc_lookup(t1, t2);
-				unref(*lhs);
-				*lhs = r;
 				t2 = force_string(t2);
 				r->vname = estrdup(t2->stptr, t2->stlen);	/* the subscript in parent array */
-
-				/* execute post-assignment routine if any */
-				if (t1->astore != NULL)
-					(*t1->astore)(t1, t2);
+				assoc_set(t1, t2, r);
+			} else if (r->type == Node_elem_new) {
+				r = force_array(r, false);
+				r->parent_array = t1;
+				t2 = force_string(t2);
+				r->vname = estrdup(t2->stptr, t2->stlen);	/* the subscript in parent array */
 			} else if (r->type != Node_var_array) {
 				t2 = force_string(t2);
 				fatal(_("attempt to use scalar `%s[\"%.*s\"]' as an array"),
 						array_vname(t1), (int) t2->stlen, t2->stptr);
-			}
+			} else
+				DEREF(t2);
 
-			DEREF(t2);
 			PUSH(r);
 			break;
 
 		case Op_subscript_lhs:
 			t2 = mk_sub(pc->sub_count);
-			t1 = POP_ARRAY();
+			t1 = POP_ARRAY(false);
 			if (do_lint && in_array(t1, t2) == NULL) {
 				t2 = force_string(t2);
 				if (pc->do_reference)
@@ -331,19 +418,22 @@ uninitialized_scalar:
 			 * be stored in SYMTAB:
 			 * 	1. Variables that don"t yet have a value (Node_var_new)
 			 * 	2. Variables that have a value (Node_var)
-			 * 	3. Values that awk code stuck into SYMTAB not related to variables (Node_value)
+			 * 	3. Values that awk code stuck into SYMTAB not related to variables (Node_val)
 			 * For 1, since we are giving it a value, we have to change the type to Node_var.
 			 * For 1 and 2, we have to step through the Node_var to get to the value.
-			 * For 3, we just us the value we got from assoc_lookup(), above.
+			 * For 3, we fatal out. This avoids confusion on things like
+			 * SYMTAB["a foo"] = 42	# variable with a space in its name?
 			 */
 			if (t1 == func_table)
 				fatal(_("cannot assign to elements of FUNCTAB"));
-			else if (   t1 == symbol_table
-				 && (   (*lhs)->type == Node_var
+			else if (t1 == symbol_table) {
+				if ((   (*lhs)->type == Node_var
 				     || (*lhs)->type == Node_var_new)) {
-				update_global_values();		/* make sure stuff like NF, NR, are up to date */
-				(*lhs)->type = Node_var;	/* in case was Node_var_new */
-				lhs = & ((*lhs)->var_value);	/* extra level of indirection */
+					update_global_values();		/* make sure stuff like NF, NR, are up to date */
+					(*lhs)->type = Node_var;	/* in case was Node_var_new */
+					lhs = & ((*lhs)->var_value);	/* extra level of indirection */
+				} else
+					fatal(_("cannot assign to arbitrary elements of SYMTAB"));
 			}
 
 			assert(set_idx == NULL);
@@ -383,16 +473,21 @@ uninitialized_scalar:
 					lintwarn(_("assignment used in conditional context"));
 					break;
 
-				case LINT_no_effect:
-					lintwarn(_("statement has no effect"));
-					break;
-
 				default:
-					cant_happen();
+					cant_happen("unexpected lint type value %d", (int) pc->lint_type);
 				}
 			}
 			break;
 
+		case Op_lint_plus:
+			// no need to check do_lint, this opcode won't
+			// be generated if that's not true
+			t1 = TOP();
+			t2 = PEEK(1);
+			if ((t1->flags & STRING) != 0 && (t2->flags & STRING) != 0)
+				lintwarn(_("operator `+' used on two string values"));
+			break;
+
 		case Op_K_break:
 		case Op_K_continue:
 		case Op_jmp:
@@ -446,37 +541,37 @@ uninitialized_scalar:
 			break;
 
 		case Op_equal:
-			r = node_Boolean[cmp_scalars(SCALAR_EQ_NEQ) == 0];
+			r = node_Boolean[cmp_scalars(SCALAR_EQ)];
 			UPREF(r);
 			REPLACE(r);
 			break;
 
 		case Op_notequal:
-			r = node_Boolean[cmp_scalars(SCALAR_EQ_NEQ) != 0];
+			r = node_Boolean[cmp_scalars(SCALAR_NEQ)];
 			UPREF(r);
 			REPLACE(r);
 			break;
 
 		case Op_less:
-			r = node_Boolean[cmp_scalars(SCALAR_RELATIONAL) < 0];
+			r = node_Boolean[cmp_scalars(SCALAR_LT)];
 			UPREF(r);
 			REPLACE(r);
 			break;
 
 		case Op_greater:
-			r = node_Boolean[cmp_scalars(SCALAR_RELATIONAL) > 0];
+			r = node_Boolean[cmp_scalars(SCALAR_GT)];
 			UPREF(r);
 			REPLACE(r);
 			break;
 
 		case Op_leq:
-			r = node_Boolean[cmp_scalars(SCALAR_RELATIONAL) <= 0];
+			r = node_Boolean[cmp_scalars(SCALAR_LE)];
 			UPREF(r);
 			REPLACE(r);
 			break;
 
 		case Op_geq:
-			r = node_Boolean[cmp_scalars(SCALAR_RELATIONAL) >= 0];
+			r = node_Boolean[cmp_scalars(SCALAR_GE)];
 			UPREF(r);
 			REPLACE(r);
 			break;
@@ -491,6 +586,7 @@ uninitialized_scalar:
 plus:
 			t1 = TOP_NUMBER();
 			r = make_number(t1->numbr + x2);
+			r->numbr = fix_nan_sign(t1->numbr, x2, r->numbr);
 			DEREF(t1);
 			REPLACE(r);
 			break;
@@ -505,6 +601,7 @@ plus:
 minus:
 			t1 = TOP_NUMBER();
 			r = make_number(t1->numbr - x2);
+			r->numbr = fix_nan_sign(t1->numbr, x2, r->numbr);
 			DEREF(t1);
 			REPLACE(r);
 			break;
@@ -621,6 +718,9 @@ mod:
 		case Op_unary_plus:
 			// Force argument to be numeric
 			t1 = TOP_NUMBER();
+			r = make_number(t1->numbr);
+			DEREF(t1);
+			REPLACE(r);
 			break;
 
 		case Op_store_sub:
@@ -641,22 +741,25 @@ mod:
 			/*
 			 * Changing something in FUNCTAB is not allowed.
 			 *
-			 * SYMTAB is a little more messy.  Three kinds of values may
-			 * be stored in SYMTAB:
+			 * SYMTAB is a little more messy.  Three possibilities for SYMTAB:
 			 * 	1. Variables that don"t yet have a value (Node_var_new)
 			 * 	2. Variables that have a value (Node_var)
-			 * 	3. Values that awk code stuck into SYMTAB not related to variables (Node_value)
+			 * 	3. Values that awk code stuck into SYMTAB not related to variables (Node_val)
 			 * For 1, since we are giving it a value, we have to change the type to Node_var.
 			 * For 1 and 2, we have to step through the Node_var to get to the value.
-			 * For 3, we just us the value we got from assoc_lookup(), above.
+			 * For 3, we fatal out. This avoids confusion on things like
+			 * SYMTAB["a foo"] = 42	# variable with a space in its name?
 			 */
 			if (t1 == func_table)
 				fatal(_("cannot assign to elements of FUNCTAB"));
-			else if (   t1 == symbol_table
-				 && (   (*lhs)->type == Node_var
+			else if (t1 == symbol_table) {
+				if ((   (*lhs)->type == Node_var
 				     || (*lhs)->type == Node_var_new)) {
-				(*lhs)->type = Node_var;	/* in case was Node_var_new */
-				lhs = & ((*lhs)->var_value);	/* extra level of indirection */
+					update_global_values();		/* make sure stuff like NF, NR, are up to date */
+					(*lhs)->type = Node_var;	/* in case was Node_var_new */
+					lhs = & ((*lhs)->var_value);	/* extra level of indirection */
+				} else
+					fatal(_("cannot assign to arbitrary elements of SYMTAB"));
 			}
 
 			unref(*lhs);
@@ -699,11 +802,18 @@ mod:
 			lhs = r_get_field(t1, & assign, false);
 			decr_sp();
 			DEREF(t1);
+			/*
+			 * N.B. We must call assign() before unref, since
+			 * we may need to copy $n values before freeing the
+			 * $0 buffer.
+			 */
+			assert(assign != NULL);
+			assign();
 			unref(*lhs);
 			r = POP_SCALAR();
 			UNFIELD(*lhs, r);
-			assert(assign != NULL);
-			assign();
+			/* field variables need the string representation: */
+			force_string(*lhs);
 		}
 			break;
 
@@ -715,7 +825,10 @@ mod:
 
 			if (t1 != *lhs) {
 				unref(*lhs);
-				*lhs = dupnode(t1);
+				if (t1->valref == 1)
+					*lhs = t1;
+				else
+					*lhs = dupnode(t1);
 			}
 
 			if (t1 != t2 && t1->valref == 1 && (t1->flags & (MALLOC|MPFN|MPZN)) == MALLOC) {
@@ -730,6 +843,9 @@ mod:
 				/* configure as a string as in make_str_node */
 				t1->flags |= (MALLOC|STRING|STRCUR);
 				t1->stfmt = STFMT_UNUSED;
+#ifdef HAVE_MPFR
+				t1->strndmode = MPFR_round_mode;
+#endif
 
 				if ((t1->flags & WSTRCUR) != 0 && (t2->flags & WSTRCUR) != 0) {
 					size_t wlen = t1->wstlen + t2->wstlen;
@@ -759,6 +875,10 @@ mod:
 			lhs = POP_ADDRESS();
 			r = TOP_SCALAR();
 			unref(*lhs);
+			if (r->type == Node_elem_new) {
+				DEREF(r);
+				r = dupnode(Nnull_string);
+			}
 			UPREF(r);
 			UNFIELD(*lhs, r);
 			REPLACE(r);
@@ -859,19 +979,19 @@ mod:
 			break;
 
 		case Op_K_delete:
-			t1 = POP_ARRAY();
+			t1 = POP_ARRAY(false);
 			do_delete(t1, pc->expr_count);
 			stack_adj(-pc->expr_count);
 			break;
 
 		case Op_K_delete_loop:
-			t1 = POP_ARRAY();
+			t1 = POP_ARRAY(false);
 			lhs = POP_ADDRESS();	/* item */
 			do_delete_loop(t1, lhs);
 			break;
 
 		case Op_in_array:
-			t1 = POP_ARRAY();
+			t1 = POP_ARRAY(false);
 			t2 = mk_sub(pc->expr_count);
 			r = node_Boolean[(in_array(t1, t2) != NULL)];
 			DEREF(t2);
@@ -890,7 +1010,7 @@ mod:
 			bool saved_end = false;
 
 			/* get the array */
-			array = POP_ARRAY();
+			array = POP_ARRAY(true);
 
 			/* sanity: check if empty */
 			num_elems = assoc_length(array);
@@ -984,14 +1104,23 @@ arrayfor:
 						(unsigned long) max_expect);
 
 			PUSH_CODE(pc);
-			r = awk_value_to_node(pc->extfunc(arg_count, & result, f));
+			awk_value_t *ef_ret = pc->extfunc(arg_count, & result, f);
+			r = awk_value_to_node(ef_ret);
 			(void) POP_CODE();
 			while (arg_count-- > 0) {
 				t1 = POP();
-				if (t1->type == Node_val)
+				if (t1->type == Node_val || t1->type == Node_elem_new)
 					DEREF(t1);
 			}
 			free_api_string_copies();
+
+			if (in_indirect_call) {
+				// pop function name off the stack
+				NODE *fname = POP();
+				DEREF(fname);
+				in_indirect_call = false;
+			}
+
 			PUSH(r);
 		}
 			break;
@@ -1036,6 +1165,10 @@ match_re:
 			if (op != Op_match_rec) {
 				decr_sp();
 				DEREF(t1);
+				if (m->type == Node_dynregex) {
+				 	DEREF(m->re_exp);
+					m->re_exp = NULL;
+				}
 			}
 			r = node_Boolean[di];
 			UPREF(r);
@@ -1061,6 +1194,7 @@ match_re:
 			NODE *f = NULL;
 			int arg_count;
 			char save;
+			NODE *function_name;
 
 			arg_count = (pc + 1)->expr_count;
 			t1 = PEEK(arg_count);	/* indirect var */
@@ -1103,6 +1237,12 @@ match_re:
 					r = the_func(arg_count);
 				str_restore(t1, save);
 
+				// Normally, setup_frame() handles getting rid of the
+				// function name.  Since we have called the builtin directly,
+				// we have to manually do this here.
+				function_name = POP();
+				DEREF(function_name);
+
 				PUSH(r);
 				break;
 			} else if (f->type != Node_func) {
@@ -1124,6 +1264,7 @@ match_re:
 					npc[1] = pc[1];
 					npc[1].func_name = fname;	/* name of the builtin */
 					npc[1].c_function = bc->c_function;
+					in_indirect_call = true;
 					ni = npc;
 					JUMPTO(ni);
 				} else
@@ -1171,6 +1312,10 @@ match_re:
 			JUMPTO(ni);	/* Op_func */
 		}
 
+		case Op_K_return_from_eval:
+			cant_happen("unexpected opcode %s", opcode2str(op));
+			break;
+
 		case Op_K_return:
 			m = POP_SCALAR();       /* return value */
 
@@ -1182,7 +1327,7 @@ match_re:
 			JUMPTO(ni);
 
 		case Op_K_getline_redir:
-			r = do_getline_redir(pc->into_var, pc->redir_type);
+			r = do_getline_redir(pc->into_var, (enum redirval) pc->redir_type);
 			PUSH(r);
 			break;
 
