@@ -1,6 +1,6 @@
 /* gawkmisc.c --- miscellaneous gawk routines that are OS specific.
- 
-   Copyright (C) 1986, 1988, 1989, 1991 - 1998, 2001 - 2004, 2011
+
+   Copyright (C) 1986, 1988, 1989, 1991 - 1998, 2001 - 2004, 2011, 2021, 2022,
    the Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
@@ -18,16 +18,18 @@
    Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 #ifdef __CYGWIN__
-#include <stdio.h>
+#ifdef __MSYS__
+#define WIN32_LEAN_AND_MEAN
 #include <windows.h>
 #include <sys/cygwin.h>
-#include <io.h>
+#endif
+#include <io.h>		/* for declaration of setmode(). */
 #endif
 
-char quote = '\'';
-char *defpath = DEFPATH;
-char *deflibpath = DEFLIBPATH;
-char envsep = ':';
+const char quote = '\'';
+const char *defpath = DEFPATH;
+const char *deflibpath = DEFLIBPATH;
+const char envsep = ':';
 
 #ifndef INVALID_HANDLE
 /* FIXME: is this value for INVALID_HANDLE correct? */
@@ -36,11 +38,11 @@ char envsep = ':';
 
 /* gawk_name --- pull out the "gawk" part from how the OS called us */
 
-char *
+const char *
 gawk_name(const char *filespec)
 {
-	char *p;
-    
+	const char *p;
+
 	/* "path/name" -> "name" */
 	p = strrchr(filespec, '/');
 	return (p == NULL ? (char *) filespec : p + 1);
@@ -272,7 +274,7 @@ os_isatty(int fd)
 {
 	return isatty(fd);
 }
- 
+
 /* files_are_same --- return true if files are identical */
 
 int
@@ -290,7 +292,13 @@ init_sockets(void)
 {
 }
 
-#ifdef __CYGWIN__
+void
+os_maybe_set_errno(void)
+{
+}
+
+// For MSYS, restore behavior of working in text mode.
+#ifdef __MSYS__
 void
 cygwin_premain0(int argc, char **argv, struct per_process *myself)
 {
